print("import: 'selenium'")
import selenium

print("import: 'selenium.common'")
import selenium.common

print("import: 'selenium.webdriver'")
import selenium.webdriver

print("import: 'selenium.webdriver.android'")
import selenium.webdriver.android

print("import: 'selenium.webdriver.chrome'")
import selenium.webdriver.chrome

print("import: 'selenium.webdriver.common'")
import selenium.webdriver.common

print("import: 'selenium.webdriver.common.html5'")
import selenium.webdriver.common.html5

print("import: 'selenium.webdriver.edge'")
import selenium.webdriver.edge

print("import: 'selenium.webdriver.firefox'")
import selenium.webdriver.firefox

print("import: 'selenium.webdriver.ie'")
import selenium.webdriver.ie

print("import: 'selenium.webdriver.opera'")
import selenium.webdriver.opera

print("import: 'selenium.webdriver.phantomjs'")
import selenium.webdriver.phantomjs

print("import: 'selenium.webdriver.remote'")
import selenium.webdriver.remote

print("import: 'selenium.webdriver.support'")
import selenium.webdriver.support

